'use strict'

var CONFIG = require('./config')

var Promise = require('bluebird')
var newify  = require('newify')
var Table   = require('alchemy/Table')
var Column  = require('alchemy/Column')
var Types  = require('alchemy/Types')

var addEntityMeta = require('./addEntityMeta')
var copy = require('./copy')

function unref(obj){
	return obj?
				JSON.parse(JSON.stringify(obj))
				:
				{}
}

function unique(arr){
    var result = []

    arr.forEach(function(value){
        if (result.indexOf(value) == -1){
            result.push(value)
        }
    })

    return result
}

function EntityError() {}
EntityError.prototype = Object.create(Error.prototype)

var EntityBase = require('classy').define({
	alias: 'entity.Base',

	properties: {},

	autoCreate : false,
	cacheEntity: true,

	statics: {

		createdByLength: 250,
		lastModifiedByLength: 250,

		init: function(){
			this.initProperties()

			if (this.prototype.table){
				this.initTable()

				if (this.autoCreate){
					this.create()
				}
			}

			if (this.prototype.cacheEntity && this.prototype.name){
				this.cache()
			}
		},

		initProperties: function(){
			var props = this.prototype.properties

			props.id = [Types.UUID, { primaryKey: true, 'default': CONFIG.uuidFn }]

			props.createdDate = [Types.Timestamp, { 'default': 'now() '}]
			props.lastModifiedDate = [Types.Timestamp, { 'default': 'now() '}]

			props.createdBy = [Types.String(this.createdByLength), {}]
			props.lastModifiedBy = [Types.String(this.createdByLength), {}]
		},

		cache: function(){
			var key   = 'entity/' + this.prototype.name
			var cache = require('./cache/entity')

			cache[key] = this
		},

		fromJSON: function(state){
			state = state || {}

			return state.extend?
						classy.define(state):
						EntityBase.extend(state)
		},

		define: function(config){
			return this.fromJSON(config || {})
		},

		initTable: function(){

			var proto = this.prototype
			var props = proto.properties

			var columns = Object.keys(props).map(function(name){

				var value = props[name]
				var args  = [name].concat(value)

				return newify(Column, args)
			})

			var args   = [proto.table, null].concat(columns)
			this.table = newify(Table, args)
		},

		toJSON: function(){
			var proto = this.prototype

		    return {
		    	properties: proto.properties,
		    	table     : proto.table,
		    	name      : proto.name
		    }
		},

		create: function(){
			var Create = this.table.create()

			return addEntityMeta(this).then(function(info){

				var Insert = info.insert
				var id     = info.id

				var Transaction = require('alchemy/Transaction')(Create, Insert)

				return Transaction
						.execute()
						.then(function(){
							return id
						})
			})
		},

		filterKeys: function(obj){
			var props = this.prototype.properties

			var result = {}

			obj && Object.keys(obj).forEach(function(key){
				if (key in obj){
					result[key] = obj[key]
				}
			})

			return result
		},

		insert: function(values){

			if (values.createdBy === undefined){
				values.createdBy = null
			}

			return this.table
					.insert()
					.values(this.filterKeys(values))
					.execute()
		},

		getBy: function(values){

			var Class = this

			return this.selectGetAllBy(values)
						.limit(1)
						.execute()
						.then(function(result){
							if (result.rows.length){
								return new Class(result.rows[0])
							}

							throw new EntityError('not found')
						})
		},

		getAllBy: function(values){
			var Class = this

			return this.selectGetAllBy()
					.execute()
					.then(function(result){
						return result.rows.map(function(row){
							new Class(row)
						})
					})
		},

		selectGetAllBy: function(values){
			var props = this.prototype.properties
			var map = {}
			var sel = this.table.select()

			Object.keys(values)
				.filter(function(key){
					return key in props
				})
				.forEach(function(key){
					var value = values[key]

					sel.where(key).is(value)
				})

			return sel
		}
	},

	init: function(values){
		this.setValues(values)
	},

	setValues: function(values){
		values = values || {}

		var props = this.properties

		var thisValues = this.values = {}

		Object.keys(props)
			.forEach(function(name){

				thisValues[name] = values[name]

				Object.defineProperty(thisValues, name, {
					set: function(value){
						values[name] = value
					},
					get: function(){
						return values[name]
					}
				})
			})
	},

	set: function(values){
		if (arguments.length == 1 && typeof values == 'object'){
			Object.keys(values).forEach(function(k){
				this.set(k, values[k])
			}, this)

			return this
		}

		var k = arguments[0]
		var v = arguments[1]

		if (k in this.properties){
			this.values[k] = v
		}

		return this
	},

	clone: function(){
		return new this.$ownClass(unref(this.values))
	},

	toJSON: function(){
		return this.values
	}
})

module.exports = EntityBase